from django.db.models import ProtectedError
from django.test import TestCase, tag
from edc_appointment.models import Appointment
from edc_lab.models import Panel
from edc_visit_tracking.constants import MISSED_VISIT, SCHEDULED

from ...constants import KEYED
from ...models import CrfMetadata, RequisitionMetadata
from ..models import CrfOne, SubjectRequisition, SubjectVisit
from .metadata_test_mixin import TestMetadataMixin


class TestDeletesMetadata(TestMetadataMixin, TestCase):
    @tag("122")
    def test_deletes_metadata_on_changed_reason_toggled(self):
        SubjectVisit.objects.create(appointment=self.appointment, reason=SCHEDULED)
        appointment = Appointment.objects.get(
            subject_identifier=self.subject_identifier,
            visit_code="2000",
        )
        obj = SubjectVisit.objects.create(appointment=appointment, reason=SCHEDULED)
        self.assertEqual(CrfMetadata.objects.filter(visit_code="2000").count(), 3)
        self.assertEqual(
            RequisitionMetadata.objects.filter(visit_code="2000").count(),
            8,
        )
        obj.reason = MISSED_VISIT
        obj.save()
        self.assertEqual(CrfMetadata.objects.filter(visit_code="2000").count(), 1)
        self.assertEqual(RequisitionMetadata.objects.filter(visit_code="2000").count(), 0)

    @tag("122")
    def test_deletes_metadata_on_changed_reason(self):
        obj = SubjectVisit.objects.create(appointment=self.appointment, reason=SCHEDULED)
        self.assertGreater(CrfMetadata.objects.all().count(), 0)
        self.assertGreater(RequisitionMetadata.objects.all().count(), 0)
        obj.reason = MISSED_VISIT
        obj.save()
        self.assertEqual(CrfMetadata.objects.all().count(), 0)
        self.assertEqual(RequisitionMetadata.objects.all().count(), 0)

    @tag("122")
    def test_deletes_metadata_on_changed_reason_adds_back_crfs_missed(self):
        SubjectVisit.objects.create(appointment=self.appointment, reason=SCHEDULED)
        appointment = Appointment.objects.get(
            subject_identifier=self.subject_identifier,
            visit_code="2000",
        )
        obj = SubjectVisit.objects.create(appointment=appointment, reason=SCHEDULED)
        self.assertGreater(CrfMetadata.objects.all().count(), 0)
        self.assertGreater(RequisitionMetadata.objects.all().count(), 0)
        obj.reason = MISSED_VISIT
        obj.save()
        self.assertEqual(CrfMetadata.objects.filter(visit_code="2000").count(), 1)
        self.assertEqual(RequisitionMetadata.objects.filter(visit_code="2000").count(), 0)

    def test_deletes_metadata_on_delete_visit(self):
        obj = SubjectVisit.objects.create(appointment=self.appointment, reason=SCHEDULED)
        self.assertGreater(CrfMetadata.objects.all().count(), 0)
        self.assertGreater(RequisitionMetadata.objects.all().count(), 0)
        obj.delete()
        self.assertEqual(CrfMetadata.objects.all().count(), 0)
        self.assertEqual(RequisitionMetadata.objects.all().count(), 0)

    @tag("100")
    def test_deletes_metadata_on_delete_visit_even_for_missed(self):
        SubjectVisit.objects.create(appointment=self.appointment, reason=SCHEDULED)
        appointment = Appointment.objects.get(
            subject_identifier=self.subject_identifier,
            visit_code="2000",
        )
        subject_visit = SubjectVisit.objects.create(appointment=appointment, reason=SCHEDULED)
        subject_visit.reason = MISSED_VISIT
        subject_visit.save()
        subject_visit.delete()
        self.assertEqual(CrfMetadata.objects.filter(visit_code="2000").count(), 0)
        self.assertEqual(RequisitionMetadata.objects.filter(visit_code="2000").count(), 0)

    @tag("101")
    def test_delete_visit_for_keyed_crf(self):
        subject_visit = SubjectVisit.objects.create(
            appointment=self.appointment, reason=SCHEDULED
        )
        self.assertGreater(CrfMetadata.objects.all().count(), 0)
        # delete
        subject_visit.delete()
        self.assertEqual(CrfMetadata.objects.all().count(), 0)
        # recreate
        subject_visit.save()
        self.assertGreater(CrfMetadata.objects.all().count(), 0)
        crf_one = CrfOne(subject_visit=subject_visit)
        crf_one.save()
        self.assertRaises(ProtectedError, subject_visit.delete)
        crf_one.delete()
        # create error condition, keyed but no model instances
        CrfMetadata.objects.all().update(entry_status=KEYED)
        subject_visit.delete()
        self.assertEqual(CrfMetadata.objects.all().count(), 0)

    @tag("101")
    def test_delete_visit_for_keyed_requisition(self):
        subject_visit = SubjectVisit.objects.create(
            appointment=self.appointment, reason=SCHEDULED
        )
        self.assertGreater(RequisitionMetadata.objects.all().count(), 0)
        panel = Panel.objects.get(name=RequisitionMetadata.objects.all()[0].panel_name)
        subject_requisition = SubjectRequisition.objects.create(
            subject_visit=subject_visit, panel=panel
        )
        RequisitionMetadata.objects.all().update(entry_status=KEYED)
        self.assertRaises(ProtectedError, subject_visit.delete)
        subject_requisition.delete()
        # create error condition, keyed but no model instances
        RequisitionMetadata.objects.all().update(entry_status=KEYED)
        subject_visit.delete()
        self.assertEqual(RequisitionMetadata.objects.all().count(), 0)



from .metadata_getter import MetadataGetter, MetadataValidator


class CrfMetadataValidator(MetadataValidator):
    pass


class CrfMetadataGetter(MetadataGetter):

    metadata_model = "edc_metadata.crfmetadata"

    metadata_validator_cls = CrfMetadataValidator



from typing import Any, List, Union

from ..constants import KEYED, REQUIRED
from ..utils import get_crf_metadata_model_cls, get_requisition_metadata_model_cls


class MetadataHelperMixin:
    """A mixin class to help with common queries against
    CRF and Requisition metadata.

    Always assumes the model instance `instance_attr` exists.
    """

    metadata_helper_instance_attr = "instance"

    @property
    def metadata_helper_instance(self) -> Any:
        if self.metadata_helper_instance_attr:
            return getattr(self, self.metadata_helper_instance_attr)
        else:
            return self

    @property
    def crf_metadata_exists(self: Any) -> bool:
        """Returns True if CRF metadata exists for this visit code."""
        return (
            get_crf_metadata_model_cls()
            .objects.filter(
                subject_identifier=self.metadata_helper_instance.subject_identifier,
                visit_schedule_name=self.metadata_helper_instance.visit_schedule_name,
                schedule_name=self.metadata_helper_instance.schedule_name,
                visit_code=self.metadata_helper_instance.visit_code,
                visit_code_sequence=self.metadata_helper_instance.visit_code_sequence,
            )
            .exists()
        )

    @property
    def crf_metadata_required_exists(self: Any) -> bool:
        """Returns True if any required CRFs for this visit code have
        not yet been keyed.
        """
        return self.get_crf_metadata_by(REQUIRED).exists()

    @property
    def crf_metadata_keyed_exists(self: Any) -> bool:
        """Returns True if any required CRFs for this visit code have
        been keyed.
        """
        return self.get_crf_metadata_by(KEYED).exists()

    def get_crf_metadata(self) -> Any:
        """Returns a queryset of crf metedata"""
        opts = dict(
            subject_identifier=self.metadata_helper_instance.subject_identifier,
            visit_schedule_name=self.metadata_helper_instance.visit_schedule_name,
            schedule_name=self.metadata_helper_instance.schedule_name,
            visit_code=self.metadata_helper_instance.visit_code,
            visit_code_sequence=self.metadata_helper_instance.visit_code_sequence,
        )
        return get_crf_metadata_model_cls().objects.filter(**opts)

    def get_crf_metadata_by(self, entry_status: Union[str, List[str]]) -> Any:
        if isinstance(entry_status, (list,)):
            opts = dict(entry_status__in=entry_status)
        else:
            opts = dict(entry_status=entry_status)
        return self.get_crf_metadata().filter(**opts)

    def get_requisition_metadata(self):
        """Returns a queryset of requisition metadata"""
        return get_requisition_metadata_model_cls().objects.filter(
            subject_identifier=self.metadata_helper_instance.subject_identifier,
            visit_schedule_name=self.metadata_helper_instance.visit_schedule_name,
            schedule_name=self.metadata_helper_instance.schedule_name,
            visit_code=self.metadata_helper_instance.visit_code,
            visit_code_sequence=self.metadata_helper_instance.visit_code_sequence,
        )

    @property
    def requisition_metadata_exists(self: Any) -> bool:
        """Returns True if requisition metadata exists for this visit code."""
        return self.get_requisition_metadata().exists()

    @property
    def requisition_metadata_required_exists(self: Any) -> bool:
        """Returns True if any required requisitions for this visit code
        have not yet been keyed.
        """
        return self.get_requisition_metadata_by(REQUIRED).exists()

    @property
    def requisition_metadata_keyed_exists(self: Any) -> bool:
        """Returns True if any required requisitions for this visit code
        have been keyed.
        """
        return self.get_requisition_metadata_by(KEYED)

    def get_requisition_metadata_by(self, entry_status: Union[str, List[str]]) -> Any:
        if isinstance(entry_status, (list,)):
            opts = dict(entry_status__in=entry_status)
        else:
            opts = dict(entry_status=entry_status)
        return self.get_requisition_metadata().filter(**opts)



from django.test import TestCase
from edc_appointment.constants import IN_PROGRESS_APPT, MISSED_APPT
from edc_visit_tracking.constants import SCHEDULED, UNSCHEDULED

from ...metadata import CreatesMetadataError
from ...metadata_updater import MetadataUpdater
from ...models import CrfMetadata, RequisitionMetadata
from ..models import SubjectVisit
from .metadata_test_mixin import TestMetadataMixin


class TestCreatesMetadata(TestMetadataMixin, TestCase):
    def test_metadata_updater_repr(self):
        obj = MetadataUpdater()
        self.assertTrue(repr(obj))

    def test_creates_metadata_on_scheduled(self):
        SubjectVisit.objects.create(appointment=self.appointment, reason=SCHEDULED)
        self.assertGreater(CrfMetadata.objects.all().count(), 0)
        self.assertGreater(RequisitionMetadata.objects.all().count(), 0)

    def test_creates_metadata_on_unscheduled(self):
        SubjectVisit.objects.create(appointment=self.appointment, reason=UNSCHEDULED)
        self.assertGreater(CrfMetadata.objects.all().count(), 0)
        self.assertGreater(RequisitionMetadata.objects.all().count(), 0)

    def test_does_not_creates_metadata_on_missed_no_crfs_missed(self):
        SubjectVisit.objects.create(appointment=self.appointment, reason=SCHEDULED)
        self.appointment_2000.appt_timing = MISSED_APPT
        self.appointment_2000.save_base(update_fields=["appt_timing"])
        self.assertEqual(
            CrfMetadata.objects.filter(visit_code=self.appointment_2000.visit_code).count(), 1
        )
        self.assertEqual(
            CrfMetadata.objects.filter(
                visit_code=self.appointment_2000.visit_code,
                model="edc_metadata.subjectvisitmissed",
            ).count(),
            1,
        )
        self.assertEqual(
            RequisitionMetadata.objects.filter(
                visit_code=self.appointment_2000.visit_code
            ).count(),
            0,
        )

    def test_unknown_reason_raises(self):
        self.appointment.appt_status = IN_PROGRESS_APPT
        self.appointment.save()
        self.appointment.refresh_from_db()
        self.assertRaises(
            CreatesMetadataError,
            SubjectVisit.objects.create,
            appointment=self.appointment,
            reason="ERIK",
        )

    def test_change_to_unknown_reason_raises(self):
        obj = SubjectVisit.objects.create(appointment=self.appointment, reason=SCHEDULED)
        obj.reason = "ERIK"
        self.assertRaises(CreatesMetadataError, obj.save)



import sys

from django.core.management.base import BaseCommand
from django.core.management.color import color_style

from edc_metadata.metadata_refresher import MetadataRefresher

style = color_style()


class Command(BaseCommand):

    help = "Update references, metadata and re-run metadatarules"

    def add_arguments(self, parser):
        parser.add_argument(
            "--all_metadata_only",
            dest="all_metadata_only",
            default="NO",
            help="YES/NO: Update metadata for all models only",
        )

    def handle(self, *args, **options) -> None:
        all_metadata_only = True if options.get("all_metadata_only", "") == "YES" else False
        metadata_refresher = MetadataRefresher(verbose=True)
        if all_metadata_only:
            sys.stdout.write("Updating metadata for all post consent models ...     \n")
            sys.stdout.write("  Note: References will not be updated;\n")
            sys.stdout.write("        Metadata rules will not be run.\n\n")
            metadata_refresher.create_or_update_metadata_for_all()
        else:
            metadata_refresher.run()



from django.apps import apps as django_apps
from django.db import models
from edc_model import models as edc_models
from edc_sites.models import CurrentSiteManager, SiteModelMixin

from ..managers import CrfMetadataManager
from .crf_metadata_model_mixin import CrfMetadataModelMixin


class CrfMetadata(CrfMetadataModelMixin, SiteModelMixin, edc_models.BaseUuidModel):

    on_site = CurrentSiteManager()

    objects = CrfMetadataManager()

    def __str__(self) -> str:
        return (
            f"CrfMeta {self.model} {self.visit_schedule_name}.{self.schedule_name}."
            f"{self.visit_code}.{self.visit_code_sequence}@{self.timepoint} "
            f"{self.entry_status} {self.subject_identifier}"
        )

    def natural_key(self) -> tuple:
        return (
            self.model,
            self.subject_identifier,
            self.schedule_name,
            self.visit_schedule_name,
            self.visit_code,
            self.visit_code_sequence,
        )

    # noinspection PyTypeHints
    natural_key.dependencies = ["sites.Site"]  # type: ignore

    @property
    def verbose_name(self) -> str:
        try:
            model = django_apps.get_model(self.model)
        except LookupError as e:
            return f"{e}. You need to regenerate metadata."
        return model._meta.verbose_name

    class Meta(CrfMetadataModelMixin.Meta, edc_models.BaseUuidModel.Meta):
        verbose_name = "Crf Metadata"
        verbose_name_plural = "Crf Metadata"
        unique_together = (
            (
                "subject_identifier",
                "visit_schedule_name",
                "schedule_name",
                "visit_code",
                "visit_code_sequence",
                "model",
            ),
        )
        indexes = [
            models.Index(
                fields=[
                    "subject_identifier",
                    "visit_schedule_name",
                    "schedule_name",
                    "visit_code",
                    "visit_code_sequence",
                    "timepoint",
                    "model",
                    "entry_status",
                    "show_order",
                ]
            )
        ]



from django.db import models

from edc_metadata.metadata_helper import MetadataHelperMixin


class MetadataHelperModelMixin(MetadataHelperMixin, models.Model):
    class Meta:
        abstract = True



from django.test import TestCase
from edc_appointment.models import Appointment
from edc_facility.import_holidays import import_holidays
from edc_lab.models.panel import Panel
from edc_reference import site_reference_configs
from edc_utils import get_utcnow
from edc_visit_schedule.site_visit_schedules import site_visit_schedules
from edc_visit_tracking.constants import SCHEDULED

from ...metadata_wrappers import (
    CrfMetadataWrapper,
    CrfMetadataWrappers,
    MetadataWrapperError,
    RequisitionMetadataWrapper,
    RequisitionMetadataWrappers,
)
from ...models import CrfMetadata, RequisitionMetadata
from ..models import CrfOne, SubjectConsent, SubjectRequisition, SubjectVisit
from ..reference_configs import register_to_site_reference_configs
from ..visit_schedule import visit_schedule


class TestMetadataWrapperObjects(TestCase):
    @classmethod
    def setUpTestData(cls):
        import_holidays()

    def setUp(self):
        self.panel_one = Panel.objects.create(name="one")

        site_visit_schedules._registry = {}
        site_visit_schedules.loaded = False
        site_visit_schedules.register(visit_schedule)

        register_to_site_reference_configs()
        site_reference_configs.register_from_visit_schedule(
            visit_models={"edc_appointment.appointment": "edc_metadata.subjectvisit"}
        )

        self.subject_identifier = "1111111"
        subject_consent = SubjectConsent.objects.create(
            subject_identifier=self.subject_identifier, consent_datetime=get_utcnow()
        )
        _, self.schedule = site_visit_schedules.get_by_onschedule_model(
            "edc_metadata.onschedule"
        )
        self.schedule.put_on_schedule(
            subject_identifier=self.subject_identifier,
            onschedule_datetime=subject_consent.consent_datetime,
        )
        self.appointment = Appointment.objects.get(
            subject_identifier=self.subject_identifier,
            visit_code=self.schedule.visits.first.code,
        )
        self.subject_visit = SubjectVisit.objects.create(
            appointment=self.appointment,
            subject_identifier=self.subject_identifier,
            reason=SCHEDULED,
        )

    def test_crf_metadata_wrapper_none(self):
        metadata_obj = CrfMetadata.objects.get(
            subject_identifier=self.subject_identifier, model="edc_metadata.crfone"
        )
        crf_metadata_wrapper = CrfMetadataWrapper(
            visit=self.subject_visit, metadata_obj=metadata_obj
        )
        self.assertEqual(crf_metadata_wrapper.model_cls, CrfOne)
        self.assertEqual(crf_metadata_wrapper.model_obj, None)
        self.assertEqual(crf_metadata_wrapper.metadata_obj, metadata_obj)
        self.assertEqual(crf_metadata_wrapper.visit, self.subject_visit)

    def test_crf_metadata_wrapper_exists(self):
        model_obj = CrfOne.objects.create(subject_visit=self.subject_visit)
        metadata_obj = CrfMetadata.objects.get(
            subject_identifier=self.subject_identifier, model="edc_metadata.crfone"
        )
        crf_metadata_wrapper = CrfMetadataWrapper(
            visit=self.subject_visit, metadata_obj=metadata_obj
        )
        self.assertEqual(crf_metadata_wrapper.model_cls, CrfOne)
        self.assertEqual(crf_metadata_wrapper.model_obj, model_obj)
        self.assertEqual(crf_metadata_wrapper.metadata_obj, metadata_obj)
        self.assertEqual(crf_metadata_wrapper.visit, self.subject_visit)

    def test_requisition_metadata_wrapper_none(self):
        metadata_obj = RequisitionMetadata.objects.get(
            subject_identifier=self.subject_identifier,
            model="edc_metadata.subjectrequisition",
            panel_name=self.panel_one.name,
        )
        requisition_metadata_wrapper = RequisitionMetadataWrapper(
            visit=self.subject_visit, metadata_obj=metadata_obj
        )
        self.assertEqual(requisition_metadata_wrapper.model_cls, SubjectRequisition)
        self.assertEqual(requisition_metadata_wrapper.model_obj, None)
        self.assertEqual(requisition_metadata_wrapper.metadata_obj, metadata_obj)
        self.assertEqual(requisition_metadata_wrapper.visit, self.subject_visit)

    def test_requisition_metadata_wrapper_exists(self):
        model_obj = SubjectRequisition.objects.create(
            subject_visit=self.subject_visit, panel=self.panel_one
        )
        metadata_obj = RequisitionMetadata.objects.get(
            subject_identifier=self.subject_identifier,
            model="edc_metadata.subjectrequisition",
            panel_name=self.panel_one.name,
        )
        requisition_metadata_wrapper = RequisitionMetadataWrapper(
            visit=self.subject_visit, metadata_obj=metadata_obj
        )
        self.assertEqual(requisition_metadata_wrapper.model_cls, SubjectRequisition)
        self.assertEqual(requisition_metadata_wrapper.model_obj, model_obj)
        self.assertEqual(requisition_metadata_wrapper.metadata_obj, metadata_obj)
        self.assertEqual(requisition_metadata_wrapper.visit, self.subject_visit)

    def test_crf_metadata_wrapper_raises_on_invalid_model(self):
        metadata_obj = CrfMetadata.objects.create(
            subject_identifier=self.subject_identifier,
            model="edc_metadata.blah",
            show_order=9999,
        )
        self.assertRaises(
            MetadataWrapperError,
            CrfMetadataWrapper,
            visit=self.subject_visit,
            metadata_obj=metadata_obj,
        )

    def test_crf_metadata_wrapper_raises_on_missing_crf_model_manager(self):
        metadata_obj = CrfMetadata.objects.create(
            subject_identifier=self.subject_identifier,
            model="edc_metadata.crfmissingmanager",
            show_order=9999,
        )
        self.assertRaises(
            MetadataWrapperError,
            CrfMetadataWrapper,
            visit=self.subject_visit,
            metadata_obj=metadata_obj,
        )

    def test_get_crfs(self):
        crf_metadata_wrappers = CrfMetadataWrappers(appointment=self.appointment)
        self.assertEqual(len(crf_metadata_wrappers.objects), 5)

    def test_get_requisitions(self):
        requisition_metadata_wrappers = RequisitionMetadataWrappers(
            appointment=self.appointment
        )
        self.assertEqual(len(requisition_metadata_wrappers.objects), 8)



from ..metadata_handler import MetadataHandler


class RequisitionMetadataHandler(MetadataHandler):

    """A class to get or create a requisition metadata
    model instance.
    """

    def __init__(self, panel=None, **kwargs):
        super().__init__(**kwargs)
        self.panel = panel

    def _create(self):
        """Returns a created metadata model instance for
        this requisition.
        """
        requisition_object = [
            r for r in self.creator.visit.all_requisitions if r.panel.name == self.panel.name
        ][0]
        return self.creator.create_requisition(requisition_object)

    @property
    def query_options(self):
        """Returns a dict of options to query the metadata model."""
        query_options = super().query_options
        query_options.update({"panel_name": self.panel.name})
        return query_options



from django.views.generic.base import ContextMixin


class RefreshMetadataViewMixin(ContextMixin):
    """
    Declare together with the edc_appointment.AppointmentViewMixin.

    """

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        # if self.appointment:
        #     pass
        return context



from collections import OrderedDict

from django.contrib.auth.models import User
from django.test import TestCase
from django.test.client import RequestFactory
from edc_appointment.models import Appointment
from edc_facility import import_holidays
from edc_form_validators.form_validator import FormValidator
from edc_lab.models import Panel
from edc_reference import site_reference_configs
from edc_utils import get_utcnow
from edc_visit_schedule import site_visit_schedules
from edc_visit_tracking.constants import SCHEDULED

from edc_metadata import site_metadata_rules
from edc_metadata.metadata_helper import MetadataHelperMixin
from edc_metadata.models import CrfMetadata, RequisitionMetadata

from ..models import SubjectConsent, SubjectVisit
from ..reference_configs import register_to_site_reference_configs
from ..visit_schedule import visit_schedule
from .test_view_mixin import MyView


class MyForm(MetadataHelperMixin, FormValidator):

    pass


class TestForm(TestCase):
    @classmethod
    def setUpTestData(cls):
        import_holidays()

    def setUp(self):
        site_metadata_rules.registry = OrderedDict()

        self.user = User.objects.create(username="erik")

        for name in ["one", "two", "three", "four", "five", "six", "seven", "eight"]:
            Panel.objects.create(name=name)

        site_visit_schedules._registry = {}
        site_visit_schedules.loaded = False
        site_visit_schedules.register(visit_schedule)

        register_to_site_reference_configs()
        site_reference_configs.register_from_visit_schedule(
            visit_models={"edc_appointment.appointment": "edc_metadata.subjectvisit"}
        )

        self.subject_identifier = "1111111"
        self.assertEqual(CrfMetadata.objects.all().count(), 0)
        self.assertEqual(RequisitionMetadata.objects.all().count(), 0)

        subject_consent = SubjectConsent.objects.create(
            subject_identifier=self.subject_identifier, consent_datetime=get_utcnow()
        )
        _, self.schedule = site_visit_schedules.get_by_onschedule_model(
            "edc_metadata.onschedule"
        )
        self.schedule.put_on_schedule(
            subject_identifier=self.subject_identifier,
            onschedule_datetime=subject_consent.consent_datetime,
        )
        self.appointment = Appointment.objects.get(
            subject_identifier=self.subject_identifier,
            visit_code=self.schedule.visits.first.code,
        )
        self.subject_visit = SubjectVisit.objects.create(
            appointment=self.appointment,
            subject_identifier=self.subject_identifier,
            reason=SCHEDULED,
        )

    def test_ok(self):
        request = RequestFactory().get("/?f=f&e=e&o=o&q=q")
        request.user = self.user
        view = MyView(request=request)
        view.appointment = self.appointment
        self.assertEqual("1000", self.appointment.visit_code)
        view.subject_identifier = self.subject_identifier
        view.kwargs = {}
        context_data = view.get_context_data()
        self.assertEqual(len(context_data.get("crfs")), 5)
        form = MyForm(cleaned_data={}, instance=view.appointment)
        self.assertTrue(form.crf_metadata_exists)
        self.assertTrue(form.crf_metadata_required_exists)
        self.assertTrue(form.requisition_metadata_exists)
        self.assertTrue(form.requisition_metadata_required_exists)
        form.validate()



from edc_reference import ReferenceModelConfig, site_reference_configs


def register_to_site_reference_configs():
    site_reference_configs.registry = {}

    reference = ReferenceModelConfig(name="edc_metadata.CrfOne", fields=["f1", "f2", "f3"])
    site_reference_configs.register(reference)

    reference = ReferenceModelConfig(name="edc_metadata.CrfTwo", fields=["f1"])
    site_reference_configs.register(reference)

    reference = ReferenceModelConfig(name="edc_metadata.CrfThree", fields=["f1"])
    site_reference_configs.register(reference)

    reference = ReferenceModelConfig(name="edc_metadata.CrfFour", fields=["f1"])
    site_reference_configs.register(reference)

    reference = ReferenceModelConfig(name="edc_metadata.CrfFive", fields=["f1"])
    site_reference_configs.register(reference)

    reference = ReferenceModelConfig(name="edc_metadata.CrfSix", fields=["f1"])
    site_reference_configs.register(reference)

    reference = ReferenceModelConfig(name="edc_metadata.CrfSeven", fields=["f1"])
    site_reference_configs.register(reference)

    reference = ReferenceModelConfig(name="edc_metadata.CrfMissingManager", fields=["f1"])
    site_reference_configs.register(reference)
